#!/usr/bin/env tclsh
#
# Scrunch Edit: a two-pane outliner for Org Mode and Markdown files
# Copyright 2021, 2022, 2023 Felix Pleșoianu <https://felix.plesoianu.ro/>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

package require Tcl 8.6
package require Tk 8.6

package require getstring
namespace import getstring::*
namespace import oo::class

set about_text "A two-pane outliner\nVersion 2.6 (6 Oct 2023)\nMIT License"
set credits_text "Made by No Time To Play\nbased on knowledge\nfrom TkDocs.com"
set site_link "https://ctrl-c.club/~nttp/toys/scrunch/"

class create MarkupParser {
	variable lines cursor head_char Heading
	
	constructor {l hc} {
		set lines $l
		
		set cursor 0
		set head_char $hc
		set Headline ""
	}
	
	method parse_section {} {
		if {$cursor >= [llength $lines]} { return "" }
		
		set body [list]
		set line [lindex $lines $cursor]
		
		while {[string index $line 0] != $head_char} {
			lappend body $line
			incr cursor
			if {$cursor >= [llength $lines]} {
				break
			} else {
				set line [lindex $lines $cursor]
			}
		}
		
		return [join $body "\n"]
	}
	
	method match_head {{level 0}} {
		if {$cursor >= [llength $lines]} { return 0 }

		set line [lindex $lines $cursor]

		for {set i 0} {$i < $level} {incr i} {
			if {[string index $line $i] != $head_char} {
				return 0
			}
		}
		
		set Heading [string trim [string range $line $i end]]
		incr cursor
		
		return 1
	}
	
	method heading {} { return $Heading }
}

proc parse_markup {parser {level 0}} {
	set subnodes [list]
	incr level
	while {[$parser match_head $level]} {
		set node [dict create \
			"text" [$parser heading] \
			"section" [$parser parse_section] \
			"children" [parse_markup $parser $level] \
		]
		lappend subnodes $node
	}
	return $subnodes
}

proc parse_file {f ext} {
	set data [split [string trimright [read $f] "\n"] "\n"]
	if {$ext eq ".md" || $ext eq ".gmi"} {
		MarkupParser create parser $data "#"
	} elseif {$ext eq ".org"} {
		MarkupParser create parser $data "*"
	} else {
		return -code break "Unknown format"
	}
	set preface [parser parse_section]
	set children [parse_markup parser]
	parser destroy
	return [list $preface $children]
}

proc write_markup {f outline head_char {level 1}} {
	foreach i $outline {
		set heading [string repeat $head_char $level]
		puts $f "$heading [dict get $i text]"
		set section [dict get $i "section"]
		if {$section ne "" && $section ne "\n"} {
			puts $f $section
		}
		set children [dict get $i "children"]
		write_markup $f $children $head_char [expr {$level+1}]
	}
}

proc write_file {f ext data} {
	set preface [dict get $data "preface"] 
	if {$preface ne "" && $preface ne "\n"} {
		puts $f $preface
	}
	if {$ext eq ".md" || $ext eq ".gmi"} {
		write_markup $f [dict get $data "outline"] "#"
	} elseif {$ext eq ".org"} {
		write_markup $f [dict get $data "outline"] "*"
	} else {
		return -code break "Unknown format"
	}
}

set file_types {
	{"All files" ".*"}
	{"Org Mode files" ".org"}
	{"Markdown / Gemini files" ".md .gmi"}
}

set node_text() ""

set file_name ""
set modified 0
set editing ""; # Which node we're currently editing, if any.

set term ""; # Current search term, if any.

namespace eval font_size {
	variable minimum 6
	variable default 11
	variable maximum 16

	variable current $default
	
	proc decrease {widget {family "Courier"}} {
		variable current
		variable minimum
		if {$current > $minimum} {
			incr current -1
		}
		$widget configure -font "$family $current"
	}

	proc increase {widget {family "Courier"}} {
		variable current
		variable maximum
		if {$current < $maximum} {
			incr current
		}
		$widget configure -font "$family $current"
	}

	proc reset {widget {family "Courier"}} {
		variable current
		variable default
		set current $default
		$widget configure -font "$family $current"
	}
}

namespace eval tk_util {
	proc pack_scrolled widget {
		set parent [winfo parent $widget]
		ttk::scrollbar $parent.scroll -orient "vertical" \
			-command "$widget yview"
		pack $widget -side "left" -fill "both" -expand 1
		$widget configure -yscrollcommand "$parent.scroll set"
		pack $parent.scroll -side "right" -fill y
	}

	proc load_text {widget content} {
		$widget delete 1.0 end
		$widget insert end $content
		$widget edit reset
		$widget edit modified 0
	}

	proc text_selection widget {
		if {[llength [$widget tag ranges sel]] > 0} {
			return [$widget get sel.first sel.last]
		} else {
			return ""
		}
	}

	proc paste_content widget {
		if {[llength [$widget tag ranges sel]] > 0} {
			$widget delete sel.first sel.last
		}
		tk_textPaste $widget
	}

	proc select_all widget {
		$widget tag remove sel 1.0 end
		$widget tag add sel 1.0 end
	}

	proc highlight_text {widget content {start "1.0"}} {
		set idx [$widget search -nocase $content $start]
		if {$idx ne ""} {
			set len [string length $content]
			set pos "$idx +$len chars"
			$widget tag remove "sel" "1.0" "end"
			$widget tag add "sel" $idx $pos
			$widget mark set "insert" $pos
			$widget see "insert"
			focus $widget
		}
	}

	proc open_line widget {
		$widget insert "insert +0 chars" "\n"
		$widget mark set "insert" "insert -1 chars"
	}

	proc fold_all widget {
		foreach i [$widget children ""] {
			$widget item $i -open 0
		}
	}

	proc unfold_all widget {
		foreach i [$widget children ""] {
			$widget item $i -open 1
		}
	}

	proc full_screen window {
		if {[wm attributes $window -fullscreen]} {
			wm attributes $window -fullscreen 0
		} else {
			wm attributes $window -fullscreen 1
		}
	}
}

wm title . "Scrunch Edit TT"
option add *tearOff 0
. configure -padx 4
# . configure -pady 4

set icon_data "
iVBORw0KGgoAAAANSUhEUgAAACAAAAAgCAMAAABEpIrGAAAADFBMVEUAAAD/AACAgID///8+Fw++
AAAABHRSTlP/AP//07BylAAAAGZJREFUOMvN01EKwCAMA9Ca3v/OcxsD20azjwnLr09sKFoTsRfA
F5EAN7AhEoABR08/u0LBmXx5CyBdpgD8idqlzpAm+RY4ULugbDN1YSB0oWCcZAd4usxB3OyfgUOB
sM2GRfTvPgAY2wicuMw+DAAAAABJRU5ErkJggg==
"
image create photo app_icon -data $icon_data
wm iconphoto . app_icon

if {[tk windowingsystem] eq "x11"} {
	ttk::style theme use "clam"
}

ttk::frame .toolbar

ttk::panedwindow .workspace -orient horizontal
ttk::frame .workspace.tree
ttk::frame .workspace.edit
.workspace add .workspace.tree -weight 1
.workspace add .workspace.edit -weight 3

set tree [ttk::treeview .workspace.tree.widget -selectmode browse -height 20]
$tree heading "#0" -text "Sections"
tk_util::pack_scrolled $tree

set editor [text .workspace.edit.text \
	-width 42 -height 24 -wrap "word" -undo 1]
font_size::reset $editor
tk_util::pack_scrolled $editor

ttk::frame .status_line
ttk::label .status_line.status -relief sunken -textvar status
ttk::sizegrip .status_line.grip

pack .toolbar -side top -pady 4
pack .workspace -side top -fill both -expand 1

pack .status_line -side bottom -fill x -pady 4
pack .status_line.status -side left -fill x -expand 1
pack .status_line.grip -side right -anchor s

ttk::button .toolbar.new -text "New" -width 8 -under 0 -command do_new
ttk::button .toolbar.bOpen -text "Open" -width 8 -under 0 -command do_open
ttk::button .toolbar.save -text "Save" -width 8 -under 0 -command do_save

ttk::separator .toolbar.sep1 -orient vertical

ttk::button .toolbar.find -text "Find" -width 8 -under 0 -command do_find
ttk::button .toolbar.again -text "Again" -width 8 -under 1 -command find_again

ttk::separator .toolbar.sep2 -orient vertical

ttk::button .toolbar.ins -text "Insert" -width 8 -under 0 -command do_ins
ttk::button .toolbar.ren -text "Rename" -width 8 -under 4 -command do_ren
ttk::button .toolbar.del -text "Delete" -width 8 -under 0 -command do_del

pack .toolbar.new -side left
pack .toolbar.bOpen -side left
pack .toolbar.save -side left

pack .toolbar.sep1 -side left -padx 4 -pady 4 -fill y

pack .toolbar.find -side left
pack .toolbar.again -side left

pack .toolbar.sep2 -side left -padx 4 -pady 4 -fill y

pack .toolbar.ins -side left
pack .toolbar.ren -side left
pack .toolbar.del -side left

. configure -menu [menu .menubar]

set m [menu .menubar.mFile]
$m add command -label "New" -command do_new -under 0 -accel "Ctrl-N"
$m add command -label "Open..." -command do_open -under 0 -accel "Ctrl-O"
$m add command -label "Save" -command do_save -under 0 -accel "Ctrl-S"
$m add separator
$m add command -label "Save as..." -command do_save_as -under 5
$m add command -label "Reload" -command do_reload -under 0 -accel "Ctrl-R"
$m add command -label "Statistics" -command show_stats -under 1 -accel "Ctrl-T"
$m add separator
$m add command -label "Quit" -command do_quit -under 0 -accel "Ctrl-Q"
.menubar add cascade -menu .menubar.mFile -label "File" -underline 0

set m [menu .menubar.edit]
$m add command -label "Undo" -command {$editor edit undo} \
	-underline 0 -accelerator "Ctrl-Z"
$m add command -label "Redo" -command {$editor edit redo} \
	-underline 0 -accelerator "Ctrl-Y"
$m add separator
$m add command -label "Cut" -command {tk_textCut $editor} \
	-underline 0 -accelerator "Ctrl-X"
$m add command -label "Copy" -command {tk_textCopy $editor} \
	-underline 1 -accelerator "Ctrl-C"
$m add command -label "Paste" -command {tk_util::paste_content $editor} \
	-underline 0 -accelerator "Ctrl-V"
$m add separator
$m add command -label "Select all" -under 0 -accel "Ctrl-A" \
	-command {tk_util::select_all $editor; break}
$m add command -label "Find..." -command do_find -under 0 -accel "Ctrl-F"
$m add command -label "Again" -command find_again -under 1 -accel "Ctrl-G"
.menubar add cascade -menu .menubar.edit -label "Edit" -underline 0

set m [menu .menubar.mFormat]
$m add command -label "Join lines" -command join_lines -under 0 -accel "Alt-J"
$m add command -label "Open line" -under 0 -accel "Alt-O" \
	-command {tk_util::open_line $editor}
$m add separator
$m add command -label "Lower case" -command lower_case -under 0 -accel "Alt-L"
$m add command -label "Title case" -command title_case -under 1 -accel "Alt-T"
$m add command -label "Upper case" -command upper_case -under 0 -accel "Alt-U"
$m add separator
$m add command -label "Prefix lines..." -command prefix_lines \
	-under 0 -accel "Alt-P"
.menubar add cascade -menu .menubar.mFormat -label "Format" -underline 3

set m [menu .menubar.view]
$m add command -label "Preface" -command {select_root $tree} \
	-underline 2 -accelerator "Ctrl-E"
$m add separator
$m add command -label "Fold all" -command {tk_util::fold_all $tree} \
	-underline 0 -accelerator "Ctrl <"
$m add command -label "Unfold all" -command {tk_util::unfold_all $tree} \
	-underline 0 -accelerator "Ctrl >"
$m add separator
$m add command -label "Bigger font" -under 0 -accel "Ctrl +" \
	-command {font_size::increase $editor}
$m add command -label "Smaller font" -under 0 -accel "Ctrl -" \
	-command {font_size::decrease $editor}
$m add command -label "Reset font" -under 0 -accel "Ctrl-0" \
	-command {font_size::reset $editor}
$m add separator
if {[llength [info commands "console"]] > 0} {
	$m add command -label "Console" -command {console show} \
		-underline 5 -accelerator "Ctrl-L"
}
$m add checkbutton -label "Full screen" -command {tk_util::full_screen .} \
	-underline 10 -accelerator "F11" -var _full_screen
.menubar add cascade -menu .menubar.view -label "View" -underline 0

set m [menu .menubar.section]
$m add command -label "Add child..." -command add_child \
	-underline 0 -accelerator "Ctrl-Ins"
$m add separator
$m add command -label "Insert after..." -command do_ins \
	-under 0 -accel "Ctrl-I"
$m add command -label "Rename..." -command do_ren -under 4 -accel "Ctrl-M"
$m add command -label "Delete" -command do_del -under 4 -accel "Ctrl-D"
$m add separator
$m add command -label "Move up" -command move_up \
	-underline 5 -accelerator "Num 8"
$m add command -label "Move right" -command move_right \
	-underline 5 -accelerator "Num 6"
$m add command -label "Move left" -command move_left \
	-underline 5 -accelerator "Num 4"
$m add command -label "Move down" -command move_down \
	-underline 5 -accelerator "Num 2"
.menubar add cascade -menu .menubar.section -label "Section" -underline 2

set m [menu .menubar.help]
$m add command -label "About" -command {alert $about_text} -under 0
$m add command -label "Credits" -command {alert $credits_text} -under 0
$m add command -label "Website" -command {open_in_app $site_link} -under 0
.menubar add cascade -menu .menubar.help -label "Help" -underline 0

wm protocol . WM_DELETE_WINDOW do_quit

bind $tree <<TreeviewSelect>> select_section
bind $editor <<Modified>> show_modified

bind . <Control-n> do_new
bind . <Control-o> do_open
bind . <Control-s> do_save
bind . <Control-r> do_reload
bind . <Control-t> show_stats
bind . <Control-q> do_quit

bind . <Command-n> do_new
bind . <Command-o> do_open
bind . <Command-s> do_save
bind . <Command-r> do_reload
bind . <Command-t> show_stats
bind . <Command-q> do_quit

# Undo, cut and copy are already bound to their usual keys by default.

bind $editor <Control-y> {$editor edit redo}
bind $editor <Control-v> {tk_util::paste_content $editor; break}
bind $editor <Control-a> {tk_util::select_all $editor; break}
bind . <Control-f> do_find
bind . <Control-g> find_again

bind $editor <Command-y> {$editor edit redo}
bind $editor <Command-v> {tk_util::paste_content $editor; break}
bind $editor <Command-a> {tk_util::select_all $editor; break}
bind . <Command-f> do_find
bind . <Command-g> find_again

bind Text <Control-o> {}
bind $editor <Alt-j> join_lines
bind $editor <Alt-o> {tk_util::open_line $editor; break}
bind $editor <Alt-l> {lower_case; break}
bind $editor <Alt-t> {title_case; break}
bind $editor <Alt-u> upper_case
bind $editor <Alt-p> prefix_lines

bind . <Control-e> {select_root $tree}
bind . <Control-comma> {tk_util::fold_all $tree}
bind . <Control-period> {tk_util::unfold_all $tree}

bind . <Command-e> {select_root $tree}
bind . <Command-comma> {tk_util::fold_all $tree}
bind . <Command-period> {tk_util::unfold_all $tree}

bind . <Control-equal> {font_size::increase $editor}
bind . <Control-minus> {font_size::decrease $editor}
bind . <Control-Key-0> {font_size::reset $editor}

bind . <Command-equal> {font_size::increase $editor}
bind . <Command-minus> {font_size::decrease $editor}
bind . <Command-Key-0> {font_size::reset $editor}

bind . <F11> {tk_util::full_screen .}

if {[llength [info commands "console"]] > 0} {
	bind . <Control-l> {console show}
	bind . <Command-l> {console show}
}

bind $tree <Insert> {add_child; break}
bind $tree <Delete> {do_del; break}

bind . <Control-Insert> add_child
bind . <Control-Delete> do_del

bind . <Command-Insert> add_child
bind . <Command-Delete> do_del

bind . <Control-i> do_ins
bind . <Control-d> do_del
bind . <Control-m> do_ren

bind . <F2> do_ren

bind . <Command-i> do_ins
bind . <Command-d> do_del
bind . <Command-m> do_ren

bind $tree <KP_8> move_up
bind $tree <KP_6>  move_right
bind $tree <KP_4> move_left
bind $tree <KP_2> move_down

try {
	bind $tree <KP_Up> move_up
	bind $tree <KP_Right> move_right
	bind $tree <KP_Left> move_left
	bind $tree <KP_Down> move_down
} on error e {
	puts stderr "Keypad arrows won't be available."
}

proc is_modified {} {
	global editor modified
	return [expr {$modified || [$editor edit modified]}]
}

proc set_modified {{state 1}} {
	global modified
	set modified $state
}

proc show_modified {} {
	global editor modified status
	if {$modified || [$editor edit modified]} {
		set status "(modified)"
	}
}

proc select_section {} {
	global editor editing node_text tree term
	
	if {[$editor edit modified]} {
		set t [string trimright [$editor get "1.0" "end"] "\n"]
		set node_text($editing) "$t\n"
		set_modified
	}
	set editing [$tree focus]
	tk_util::load_text $editor $node_text($editing)
	if {$term ne ""} {
		tk_util::highlight_text $editor $term
	}
}

proc do_new {} {
	global status tree editor file_name node_text

	if {[is_modified]} {
		set answer [tk_messageBox -parent . \
			-type "yesno" -icon "question" \
			-title "Scrunch Edit TT" \
			-message "New outline?" \
			-detail "Outline is unsaved.\nStart another?"]
		if {!$answer} {
			set status "New outline canceled."
			return
		}
	}
	$tree delete [$tree children ""]
	wm title . "Scrunch Edit TT"
	$editor delete "1.0" "end"
	set file_name ""
	array unset node_text
	set node_text() ""
	$editor edit reset
	$editor edit modified 0
	set_modified 0
	set status [clock format [clock seconds]]
}

proc do_open {} {
	global status file_types file_name
	
	if {[is_modified]} {
		set answer [tk_messageBox -parent . \
			-type "yesno" -icon "question" \
			-title "Scrunch Edit TT" \
			-message "Open another outline?" \
			-detail "Outline is unsaved.\nOpen another?"]
		if {!$answer} {
			set status "Opening canceled."
			return
		}
	}
	
	if {[tk windowingsystem] == "x11" && [auto_execok "zenity"] ne ""} {
		set choice [open_with_zenity $file_name]
	} else {
		set choice [tk_getOpenFile -parent . \
			-title "Open existing outline" \
			-initialdir [file_dir $file_name] \
			-filetypes $file_types]
	}
	
	if {[string length $choice] == 0} {
		set status "Opening canceled."
	} elseif {![file isfile $choice]} {
		tk_messageBox -parent . \
			-type "ok" -icon "error" \
			-title "Scrunch Edit TT" \
			-message "Error opening file" \
			-detail "File not found: $choice"
	} elseif {[load_file $choice]} {  
		set file_name $choice
	}
}

proc open_with_zenity file_name {
	try {
		if {$file_name ne ""} {
			return [exec zenity --file-selection \
				--title "Open existing outline" \
				--filename $file_name \
				--file-filter "All files | *" \
				--file-filter "Org Mode files | *.org" \
				--file-filter "Markdown / Gemini | *.md *.gmi"]
		} else {
			return [exec zenity --file-selection \
				--title "Open existing outline" \
				--file-filter "All files | *" \
				--file-filter "Org Mode files | *.org" \
				--file-filter "Markdown / Gemini | *.md *.gmi"]
		}
	} trap CHILDSTATUS {results options} {
		return ""
	}
}

proc load_file full_path {
	global tree editor node_text modified status
	
	set fn [file tail $full_path]
	set ext [file extension $fn]
	try {
		set f [open $full_path]
		lassign [parse_file $f $ext] node_text() outline 
		$tree delete [$tree children ""]
		load_outline $outline
		tk_util::load_text $editor $node_text()
		set modified 0
		set status "Opened $fn"
		wm title . "$fn | Scrunch Edit TT"
		return 1
	} on break e {
		tk_messageBox -parent . \
			-type "ok" -icon "warning" \
			-title "Scrunch Edit TT" \
			-message "Error parsing outline" \
			-detail $e
		return 0
	} on error e {
		tk_messageBox -parent . \
			-type "ok" -icon "error" \
			-title "Scrunch Edit TT" \
			-message "Error opening file" \
			-detail $e
		return 0
	} finally {
		close $f
	}
}

proc load_outline {data {item ""}} {
	global tree node_text

	foreach i $data {
		set t [dict get $i "text"]
		set added [$tree insert $item "end" -text $t -open 1]
		set node_text($added) [dict get $i "section"]
		load_outline [dict get $i "children"] $added
	}
}

proc file_dir name {
	if {$name ne ""} {
		return [file dirname $name]
	} else {
		return [pwd]
	}
}

proc do_save {} {
	global file_name
	if {$file_name eq ""} {
		do_save_as
	} else {
		save_file $file_name
	}
}

proc do_save_as {} {
	global file_name file_types status
	
	if {[tk windowingsystem] == "x11" && [auto_execok "zenity"] ne ""} {
		set choice [save_with_zenity $file_name]
	} else {
		set choice [tk_getSaveFile -parent . \
			-title "Save outline as..." \
			-initialdir [file_dir $file_name] \
			-filetypes $file_types]
	}
	if {[string length $choice] == 0} {
		set status "Save canceled."
	} elseif {[save_file $choice]} {
		set file_name $choice
	}
}

proc save_with_zenity file_name {
	try {
		if {$file_name ne ""} {
			return [exec zenity --file-selection \
				--title "Save outline as..." \
				 --save --confirm-overwrite \
				--filename $file_name \
				--file-filter "All files | *" \
				--file-filter "Org Mode files | *.org" \
				--file-filter "Markdown / Gemini | *.md *.gmi"]
		} else {
			return [exec zenity --file-selection \
				--title "Save outline as..." \
				 --save --confirm-overwrite \
				--file-filter "All files | *" \
				--file-filter "Org Mode files | *.org" \
				--file-filter "Markdown / Gemini | *.md *.gmi"]
		}
	} trap CHILDSTATUS {results options} {
		return ""
	}
}

proc save_file full_path {
	global node_text editing editor status

	if {[$editor edit modified]} {
		set t [string trimright [$editor get "1.0" "end"] "\n"]
		set node_text($editing) "$t\n"
		set_modified
	}
	set fn [file tail $full_path]
	set ext [file extension $fn]
	set data [dict create \
		"preface" $node_text() \
		"outline" [unload_outline] \
	]
	try {
		set f [open $full_path "w"]
		write_file $f $ext $data
		flush $f
		set_modified 0
		set status "Saved $fn"
		wm title . "$fn | Scrunch Edit TT"
		return 1
	} on break e {
		tk_messageBox -parent . \
			-type "ok" -icon "error" \
			-title "Scrunch Edit TT" \
			-message "Error writing outline" \
			-detail $e
		return 0
	} on error e {
		tk_messageBox -parent . \
			-type "ok" -icon "error" \
			-title "Scrunch Edit TT" \
			-message "Error saving file" \
			-detail $e
		return 0
	} finally {
		close $f
	}
}

proc unload_outline {{item ""}} {
	global tree node_text
	set outline [list]
	foreach i [$tree children $item] {
		set child [dict create \
			"text" [$tree item $i -text] \
			"section" $node_text($i) \
			"children" [unload_outline $i] \
		]
		lappend outline $child
	}
	return $outline
}

proc do_reload {} {
	global file_name status
	if {$file_name eq ""} {
		tk_messageBox -parent . \
			-type "ok" -icon "warning" \
			-title "Scrunch Edit TT" \
			-message "Can't reload." \
			-detail "The outline was never saved."
	} else {
		set answer [tk_messageBox -parent . \
			-type "yesno" -icon "question" \
			-title "Scrunch Edit TT" \
			-message "Reload outline?" \
			-detail "Reload last save?"]
		if {$answer eq "yes"} {
			load_file $file_name
		} else {
			set status "Reloading canceled."
		}
	}
}

proc show_stats {} {
	global node_text
	set names [array names node_text]
	set sec [llength $names]
	set chars 0
	foreach i $names {
		incr chars [string length $node_text($i)]
	}
	set average [expr {$chars / $sec}]
	set stats "Sections: $sec\nCharacters: $chars\nAverage: $average"
	tk_messageBox -parent . \
		-type "ok" -icon "info" \
		-title "Scrunch Edit TT" \
		-message "Outline statistics" \
		-detail $stats
}

proc do_quit {} {
	if {[is_modified]} {
		set answer [tk_messageBox -parent . \
			-type "yesno" -icon "question" \
			-title "Scrunch Edit TT" \
			-message "Quit Scrunch Edit?" \
			-detail "Outline is unsaved.\nQuit anyway?"]
	} else {
		set answer "yes"
	}
	if {$answer eq "yes"} {
		destroy .
	}
}

proc do_find {} {
	global term editor status tree
	if {[llength [info commands "search"]] > 0} {
		rename search {}
	}
	set term [tk_util::text_selection $editor]
	set ret [tk_getString .gs answer "Search pattern:" \
		-title "Find" -entryoptions "-textvar term"]
	if {!$ret} {
		set status "Search canceled."
		return
	}
	coroutine search run_search $answer [all_item_ids $tree]
	set term $answer; # Possibly redundant
	step_search
}

# Return a flat list of item IDs present in the tree.
proc all_item_ids {tree {item ""}} {
	set items [list]
	foreach i [$tree children $item] {
		lappend items $i {*}[all_item_ids $tree $i]
	}
	return $items
}

proc find_again {} {
	if {[llength [info commands "search"]] == 0} {
		do_find
	} else {
		step_search
	}
}

proc step_search {} {
	global term status tree
	set next [search]
	if {$next eq "\n"} {
		set term ""
		set status "Nothing found."
	} elseif {$next eq ""} {
		select_preface
	} else {
		$tree selection set $next
		$tree focus $next
		$tree see $next
	}
}

proc run_search {term items} {
	global node_text tree
	
	yield "\n"
	
	set term [string tolower $term]
	set section [string tolower $node_text()]
	if {[string first $term $section] > -1} {
	 	yield ""
	}
	foreach i $items {
		if {![$tree exists $i]} {
			# It might have been deleted.
			continue
		}
		set node [string tolower [$tree item $i -text]]
		set section [string tolower $node_text($i)]
		if {[string first $term $node] > -1} {
			yield $i
		} elseif {[string first $term $section] > -1} {
			yield $i
		}
	}
	
	return "\n"
}

proc join_lines {} {
	global editor status
	set t [tk_util::text_selection $editor]
	if {[string length $t] > 0} {
		set t [string map {"\n" " "} $t]
		$editor replace sel.first sel.last $t
	} else {
		set status "Nothing selected."
	}
}

proc lower_case {} {
	global editor status
	set t [tk_util::text_selection $editor]
	if {[string length $t] > 0} {
		$editor replace sel.first sel.last [string tolower $t]
	} else {
		set status "Nothing selected."
	}
}

proc title_case {} {
	global editor status
	set sel [tk_util::text_selection $editor]
	if {[string length $sel] > 0} {
		$editor replace sel.first sel.last [string totitle $sel]
	} else {
		set status "Nothing selected."
	}
}

proc upper_case {} {
	global editor status
	set sel [tk_util::text_selection $editor]
	if {[string length $sel] > 0} {
		$editor replace sel.first sel.last [string toupper $sel]
	} else {
		set status "Nothing selected."
	}
}

proc select_root widget {
	$widget selection set ""
	$widget focus ""
	# Don't call [select_section] here, it's implied by [selection set]
}

proc add_child {} {
	global tree status
	set ret [tk_getString .gs answer "New section title:" \
		-title "Add section"]
	if {$ret} {
		set focus [$tree focus]
		set child [$tree insert $focus "end" -text $answer]
		set_up_node $child
	} else {
		set status "Canceled adding section."
	}
}

proc set_up_node node {
	global node_text tree
	set node_text($node) ""
	$tree selection set $node
	$tree focus $node
	$tree see $node
	set_modified
	show_modified
}

proc do_ins {} {
	global tree status
	set ret [tk_getString .gs answer "New section title:" \
		-title "Add section"]
	if {$ret} {
		set focus [$tree focus]
		if {$focus eq ""} {
			set parent ""
			set position "end"
		} else {
			set parent [$tree parent $focus]
			set position [expr {[$tree index $focus] + 1}]
		}
		set child [$tree insert $parent $position -text $answer]
		set_up_node $child
	} else {
		set status "Canceled adding section."
	}
}

proc do_ren {} {
	global tree status _ren_from
	set focus [$tree focus]
	if {$focus eq ""} {
		set status "Can't rename preface."
	} else {
		set _ren_from [$tree item $focus -text]
		set ret [tk_getString .gs answer "New section title:" \
			-title "Rename section" \
			-entryoptions "-textvar _ren_from"]
		if {$ret} {
			$tree item $focus -text $answer
			set_modified
			show_modified
		} else {
			set status "Canceled renaming section."
		}
	}
}

proc do_del {} {
	global tree status node_text
	set focus [$tree focus]
	if {$focus eq ""} {
		set status "Can't delete preface."
		return
	} elseif {[is_empty_section $focus]} {
		set answer "yes"
	} else {
		set answer [tk_messageBox -parent . \
			-type "yesno" -icon "question" \
			-title "Scrunch Edit TT" \
			-message "Delete section?" \
			-detail "Section isn't empty.\nDelete anyway?"]
	}
	if {$answer eq "yes"} {
		$tree selection set [$tree next $focus]
		$tree focus [$tree next $focus]
		$tree delete $focus
		unset node_text($focus)
		set_modified
		show_modified
	} else {
		set status "Deletion canceled."
	}
}

# Deliberately doesn't check the editor state.
proc is_empty_section id {
	global tree node_text
	if {$node_text($id) ne ""} {
		return 0
	} elseif {[llength [$tree children $id]] > 0} {
		return 0
	} else {
		return 1
	}
}

proc move_left {} {
	global tree status
	set focus [$tree focus]
	if {$focus eq ""} {
		set status "Can't move preface."
		return
	}
	set parent [$tree parent $focus]
	if {$parent eq ""} {
		set status "Section is at top level."
		return
	}
	set index [$tree index $parent]
	$tree move $focus [$tree parent $parent] [expr {$index + 1}]
	set_modified
	show_modified
}

proc move_down {} {
	global tree status
	set focus [$tree focus]
	if {$focus eq ""} {
		set status "Can't move preface."
		return
	}
	set index [$tree index $focus]
	$tree move $focus [$tree parent $focus] [expr {$index + 1}]
	set_modified
	show_modified
}

proc move_up {} {
	global tree status
	set focus [$tree focus]
	if {$focus eq ""} {
		set status "Can't move preface."
		return
	}
	set index [$tree index $focus]
	$tree move $focus [$tree parent $focus] [expr {$index - 1}]
	set_modified
	show_modified
}

proc move_right {} {
	global tree status
	set focus [$tree focus]
	if {$focus eq ""} {
		set status "Can't move preface."
		return
	}
	set prev [$tree prev $focus]
	if {$prev eq ""} {
		set status "No previous section."
		return
	}
	$tree move $focus $prev "end"
	$tree see $focus
	set_modified
	show_modified
}

proc prefix_lines {} {
	global editor status
	set sel [tk_util::text_selection $editor]
	if {$sel eq ""} {
		set status "Nothing selected."
		return
	}
	set ret [tk_getString .gs answer "Prefix selected lines with:"]
	if {!$ret} {
		set status "Search canceled."
		return
	}
	set lines [split $sel "\n"]
	set result [list]
	foreach l $lines {
		lappend result $answer$l
	}
	$editor replace sel.first sel.last [join $result "\n"]
}

proc alert message {
	tk_messageBox -parent . \
		-type "ok" -icon "info" \
		-title "Scrunch Edit TT" \
		-message $message
}

proc open_in_app link {
	global status
	if {[auto_execok "xdg-open"] ne ""} {
		catch {exec "xdg-open" $link &}
	} elseif {[auto_execok "open"] ne ""} {
		catch {exec "open" $link &}
	} elseif {[auto_execok "start"] ne ""} {
		catch {exec "start" $link &}
	} else {
		set status "Can't open website."
	}
}

set status [clock format [clock seconds]]

if {[llength $argv] > 0} {
	set fn [lindex $argv 0]
	if {![file exists $fn]} {
		set file_name [file normalize $fn]
		set fn [file tail $file_name]
		wm title . "$fn | Scrunch Edit TT"
	} elseif {[load_file $fn]} {
		set file_name [file normalize $fn]
	}
}
